# -*- coding: utf-8 -*-
"""
Source watcher
==============

Watcher is almost *isolated* from command line code because it runs in an
infinite loop, so note that handlers directly output some informations on a
``logging.logger``.

"""
import os
import logging


from watchdog.events import PatternMatchingEventHandler

from .compiler import SassCompileHelper
from .exceptions import BoussoleBaseException
from .finder import ScssFinder
from .utils import match_path


class SassLibraryEventHandler(object):
    """
    Watch mixin handler for library sources

    Handler does not compile source which triggered an event,
    only its parent dependencies. Because libraries are not intended to be
    compiled.

    Args:
        settings (boussole.conf.model.Settings): Project settings.
        inspector (boussole.inspector.ScssInspector): Inspector instance.

    Attributes:
        settings (boussole.conf.model.Settings): Filled from argument.
        logger (logging.Logger): Boussole logger.
        inspector (boussole.inspector.ScssInspector): Filled from argument.
        finder (boussole.finder.ScssFinder): Finder instance.
        compiler (boussole.compiler.SassCompileHelper): Sass compile helper
            object.
        compilable_files (dict): Pair of (source path, destination path) to
            compile. Automatically update from ``index()`` method.
        source_files (list): List of source path to compile. Automatically
            update from ``index()`` method.
        _event_error (bool): Internal flag setted to ``True`` if error has
            occured within an event. ``index()`` will reboot it to ``False``
            each time a new event occurs.
    """
    SUPPORTED_EVENTS = (
        "moved",
        "created",
        "modified",
        "deleted",
    )

    def __init__(self, settings, inspector, *args, **kwargs):
        self.settings = settings
        self.inspector = inspector

        self.logger = logging.getLogger("boussole")
        self.finder = ScssFinder()
        self.compiler = SassCompileHelper()

        self.compilable_files = {}
        self.source_files = []
        self._event_error = False

        super(SassLibraryEventHandler, self).__init__(*args, **kwargs)

    def is_valid_event(self, event):
        """
        Check if given event is valid event for index method.

        An event is considered valid if event type is supported (from
        ``SassLibraryEventHandler.SUPPORTED_EVENTS``) and file is allowed (from
        method ``ImportPathsResolver.is_allowed_source``).

        Args:
            event (watchdog.events.FileSystemEvent): Watchdog file system event.

        Returns:
            bool: True if valid, else False.
        """
        # Don't continue for non supported event
        if event.event_type not in self.SUPPORTED_EVENTS or event.is_directory:
            return False

        # We commonly care only about destination path, but it is missing from
        # some event where the source path is the only one available and so the
        # legit path to look at
        target_path = event.src_path
        if hasattr(event, "dest_path"):
            target_path = event.dest_path

        # Don't continue for files we don't care about
        if not self.inspector.is_allowed_source(target_path):
            return False

        return True

    def index(self, event):
        """
        Reset inspector buffers and index project sources dependencies.

        This have to be executed each time an event occurs.

        Args:
            event (watchdog.events.FileSystemEvent): Watchdog file system event.

        Returns:
            bool: True if allowed, else False.

        Note:
            If a Boussole exception occurs during operation, it will be catched
            and an error flag will be set to ``True`` so event operation will
            be stopped without blocking or breaking watchdog observer.
        """
        self._event_error = False

        # Don't continue for non valid event
        if not self.is_valid_event(event):
            return

        try:
            compilable_files = self.finder.mirror_sources(
                self.settings.SOURCES_PATH,
                targetdir=self.settings.TARGET_PATH,
                excludes=self.settings.EXCLUDES
            )
            self.compilable_files = dict(compilable_files)
            self.source_files = self.compilable_files.keys()

            # Init inspector and do first inspect
            self.inspector.reset()
            self.inspector.inspect(
                *self.source_files,
                library_paths=self.settings.LIBRARY_PATHS
            )
        except BoussoleBaseException as e:
            self._event_error = True
            self.logger.error(str(e))

    def compile_source(self, sourcepath):
        """
        Compile source to its destination

        Check if the source is eligible to compile (not partial and allowed
        from exclude patterns)

        Args:
            sourcepath (string): Sass source path to compile to its
                destination using project settings.

        Returns:
            tuple or None: A pair of (sourcepath, destination), if source has
            been compiled (or at least tried). If the source was not
            eligible to compile, return will be ``None``.
        """
        relpath = os.path.relpath(sourcepath, self.settings.SOURCES_PATH)

        conditions = {
            "sourcedir": None,
            "nopartial": True,
            "exclude_patterns": self.settings.EXCLUDES,
            "excluded_libdirs": self.settings.LIBRARY_PATHS,
        }
        if self.finder.match_conditions(sourcepath, **conditions):
            destination = self.finder.get_destination(
                relpath,
                targetdir=self.settings.TARGET_PATH
            )

            self.logger.debug("Compile: {}".format(sourcepath))
            success, message = self.compiler.safe_compile(
                self.settings,
                sourcepath,
                destination
            )

            if success:
                self.logger.info("Output: {}".format(message))
            else:
                self.logger.error(message)

            return sourcepath, destination

        return None

    def compile_dependencies(self, sourcepath, include_self=False):
        """
        Register source(s) for compile and possibly its dependencies.

        Args:
            sourcepath (string): Sass source path to compile to its
                destination using project settings.

        Keyword Arguments:
            include_self (bool): If ``True`` the given sourcepath is added to
                items to compile, else only its dependencies are compiled.
        """
        items = self.inspector.parents(sourcepath)

        # Also add the current event related path
        if include_self:
            items.add(sourcepath)

        return filter(None, [self.compile_source(item) for item in items])

    def on_any_event(self, event):
        """
        Catch-all event handler (moved, created, deleted, changed).

        Before any event, we index project to have the right and current
        dependencies map.

        Args:
            event: Watchdog event ``watchdog.events.FileSystemEvent``.
        """
        self.index(event)

    def on_moved(self, event):
        """
        Called when a file or a directory is moved or renamed.

        Many editors don't directly change a file, instead they make a
        transitional file like ``*.part`` then move it to the final filename.

        Args:
            event: Watchdog event, either ``watchdog.events.DirMovedEvent`` or
                ``watchdog.events.FileModifiedEvent``.
        """
        if not self._event_error:
            # We are only interested for final file, not transitional file
            # from editors (like *.part)
            pathtools_options = {
                "included_patterns": set(self.patterns),
                "excluded_patterns": set(self.ignore_patterns),
                "case_sensitive": self.case_sensitive,
            }
            # Apply pathtool matching on destination since Watchdog only
            # automatically apply it on source
            if match_path(event.dest_path, **pathtools_options):
                self.logger.info(
                    "Change detected from a move on: {}".format(event.dest_path)
                )
                self.compile_dependencies(event.dest_path)

    def on_created(self, event):
        """
        Called when a new file or directory is created.

        Todo:
            This should be also used (extended from another class?) to watch
            for some special name file (like ".boussole-watcher-stop" create to
            raise a KeyboardInterrupt, so we may be able to unittest the
            watcher (click.CliRunner is not able to send signal like CTRL+C
            that is required to watchdog observer loop)

        Args:
            event: Watchdog event, either ``watchdog.events.DirCreatedEvent``
                or ``watchdog.events.FileCreatedEvent``.
        """
        if not self._event_error:
            self.logger.info(
                "Change detected from a create on: {}".format(event.src_path)
            )

            self.compile_dependencies(event.src_path)

    def on_modified(self, event):
        """
        Called when a file or directory is modified.

        Args:
            event: Watchdog event, ``watchdog.events.DirModifiedEvent`` or
                ``watchdog.events.FileModifiedEvent``.
        """
        if not self._event_error:
            self.logger.info(
                "Change detected from an edit on: {}".format(event.src_path)
            )

            self.compile_dependencies(event.src_path)

    def on_deleted(self, event):
        """
        Called when a file or directory is deleted.

        Args:
            event: Watchdog event, ``watchdog.events.DirDeletedEvent`` or
                ``watchdog.events.FileDeletedEvent``.
        """
        if not self._event_error:
            self.logger.info(
                "Change detected from deletion of: {}".format(event.src_path)
            )
            # Never try to compile the deleted source
            self.compile_dependencies(event.src_path, include_self=False)


class SassProjectEventHandler(SassLibraryEventHandler):
    """
    Watch mixin handler for project sources.

    Source that trigger event is compiled (if eligible) with its dependencies.

    Warning:
        DO NOT use this handler to watch libraries, there is a risk for
        compiler trying to compile their sources in a wrong directory.
    """
    def compile_dependencies(self, sourcepath, include_self=True):
        """
        Same as inherit method but the default value for keyword argument
        ``ìnclude_self`` is ``True``.
        """
        return super(SassProjectEventHandler, self).compile_dependencies(
            sourcepath,
            include_self=include_self
        )


class WatchdogLibraryEventHandler(SassLibraryEventHandler,
                                  PatternMatchingEventHandler):
    """
    Watchdog event handler for library sources
    """
    pass


class WatchdogProjectEventHandler(SassProjectEventHandler,
                                  PatternMatchingEventHandler):
    """
    Watchdog event handler for project sources.

    Warning:
        DO NOT use this handler to watch libraries, there is a risk for
        compiler trying to compile their sources in a wrong directory.
    """
    pass



# -*- coding: utf-8 -*-
"""
Utils
=====

"""
from pathlib import PureWindowsPath, PurePosixPath


def match_path(path, included_patterns, excluded_patterns, case_sensitive):
    """
    Matches a pathname against a set of acceptable and ignored patterns.

    Inspired from ``watchdog.utils.patterns_match_path`` which was inspired from
    deprecated ``pathtools.patterns.match_path``.

    Arguments:
        path (string): A pathname which will be matched against a pattern.
        included_patterns (set): Allow filenames matching wildcard patterns
            specified in this list. If no pattern is specified, the function
            treats the pathname as a match_path.
        excluded_patterns (set): Ignores filenames matching wildcard patterns
            specified in this list. If no pattern is specified, the function
            treats the pathname as a match_path.
        case_sensitive (boolean): ``True`` if matching should be
            case-sensitive; ``False`` otherwise.

    Raises:
        ValueError: if included patterns and excluded patterns contain the
            same pattern.
    Returns:
        boolean: ``True`` if the pathname matches; ``False`` otherwise.
    """
    if case_sensitive:
        path = PurePosixPath(path)
    else:
        included_patterns = {pattern.lower() for pattern in included_patterns}
        excluded_patterns = {pattern.lower() for pattern in excluded_patterns}
        path = PureWindowsPath(path)

    common_patterns = included_patterns & excluded_patterns

    if common_patterns:
        msg = "conflicting patterns '{}' included and excluded"
        raise ValueError(msg.format(common_patterns))

    return (
        any(path.match(p) for p in included_patterns) and
        not any(path.match(p) for p in excluded_patterns)
    )



"""
Main entrance to commandline actions
"""
import click

from ..logs import init_logger

from .compile import compile_command
from .startproject import startproject_command
from .version import version_command
from .watch import watch_command


# Help alias on "-h" argument
CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])

# Default logger conf
BOUSSOLE_LOGGER_CONF = ("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", None)


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option(
    "-v",
    "--verbose",
    type=click.IntRange(min=0, max=5),
    default=4,
    metavar="INTEGER",
    help=(
        "An integer between 0 and 5, where '0' make a totaly "
        "silent output and '5' set level to DEBUG (the most verbose "
        "level). Default to '4' (Info level)."
    )
)
@click.pass_context
def cli_frontend(ctx, verbose):
    """
    Boussole is a commandline interface to build Sass projects using libsass.

    Every project will need a settings file containing all needed settings to
    build it.
    """
    printout = True
    if verbose == 0:
        verbose = 1
        printout = False

    # Verbosity is the inverse of logging levels
    levels = [item for item in BOUSSOLE_LOGGER_CONF]
    levels.reverse()
    # Init the logger config
    root_logger = init_logger(levels[verbose], printout=printout)

    # Init the default context that will be passed to commands
    ctx.obj = {
        "verbosity": verbose,
        "logger": root_logger,
    }


# Attach commands methods to the main grouper
cli_frontend.add_command(version_command, name="version")
cli_frontend.add_command(compile_command, name="compile")
cli_frontend.add_command(watch_command, name="watch")
cli_frontend.add_command(startproject_command, name="startproject")



# -*- coding: utf-8 -*-
import click
import logging
import os

from ..compiler import SassCompileHelper
from ..conf.discovery import Discover
from ..conf.json_backend import SettingsBackendJson
from ..conf.yaml_backend import SettingsBackendYaml
from ..exceptions import BoussoleBaseException
from ..finder import ScssFinder
from ..project import ProjectBase


@click.command(
    "watch",
    short_help="Compile Sass project sources to CSS."
)
@click.option(
    "--backend",
    metavar="STRING",
    type=click.Choice(["json", "yaml"]),
    help="Settings format name",
    default="json"
)
@click.option(
    "--config",
    default=None,
    metavar="PATH",
    help="Path to a Boussole config file",
    type=click.Path(exists=True)
)
@click.pass_context
def compile_command(context, backend, config):
    """
    Compile Sass project sources to CSS
    """
    logger = logging.getLogger("boussole")
    logger.info("Building project")

    # Discover settings file
    try:
        discovering = Discover(backends=[SettingsBackendJson,
                                         SettingsBackendYaml])
        config_filepath, config_engine = discovering.search(
            filepath=config,
            basedir=os.getcwd(),
            kind=backend
        )

        project = ProjectBase(backend_name=config_engine._kind_name)
        settings = project.backend_engine.load(filepath=config_filepath)
    except BoussoleBaseException as e:
        logger.critical(str(e))
        raise click.Abort()

    logger.debug("Settings file: {} ({})".format(
                 config_filepath, config_engine._kind_name))
    logger.debug("Project sources directory: {}".format(
                 settings.SOURCES_PATH))
    logger.debug("Project destination directory: {}".format(
                 settings.TARGET_PATH))
    logger.debug("Exclude patterns: {}".format(
                 settings.EXCLUDES))

    # Find all sources with their destination path
    try:
        compilable_files = ScssFinder().mirror_sources(
            settings.SOURCES_PATH,
            targetdir=settings.TARGET_PATH,
            excludes=settings.EXCLUDES
        )
    except BoussoleBaseException as e:
        logger.error(str(e))
        raise click.Abort()

    # Build all compilable stylesheets
    compiler = SassCompileHelper()
    errors = 0
    for src, dst in compilable_files:
        logger.debug("Compile: {}".format(src))

        output_opts = {}
        success, message = compiler.safe_compile(settings, src, dst)

        if success:
            logger.info("Output: {}".format(message), **output_opts)
        else:
            errors += 1
            logger.error(message)

    # Ensure correct exit code if error has occured
    if errors:
        raise click.Abort()



# -*- coding: utf-8 -*-
"""
.. _Sass Reference:
    http://sass-lang.com/documentation/file.SASS_REFERENCE.html#import

Parser
======

Parser is in charge to find every ``@import`` rules in given Sass content.

It has been built following `Sass Reference`_ about ``@import`` rule.

Note:
    Sass indented syntax parser has a flaw which cause multiline comments to be
    incorrectly parsed. If you have commented ``@import`` in multiline comments
    they will be matched as true importations to resolve and inspect, it can leads
    to some errors.

    The only way is to not use multiline comments when you use the Sass index
    syntax.
"""
import re

from .exceptions import InvalidImportRule


class ScssImportsParser(object):
    """
    SCSS parser to find ``@import`` rules.

    This does not support the old Sass syntax (also known as "indented
    syntax").

    It's a mixin, meaning without own ``__init__`` method so it's should be
    safe enough to inherit it from another class.

    Attributes:
        REGEX_IMPORT_RULE: Compiled regex used to find ``@import`` rules.
        REGEX_COMMENTS: Compiled regex used to find and remove comments.
    """
    syntax = "scss"
    REGEX_IMPORT_RULE = re.compile(r'@import\s*(url)?\s*\(?([^;]+?)\)?;',
                                   re.IGNORECASE)
    # Second part (for singleline comment) contain a negative lookbehind
    # assertion to avoid to match on url protocole (http://) which cause issues
    # in parsing
    REGEX_COMMENTS = re.compile(r'(/\*.*?\*/)|((?<!(:))//.*?(\n|$))',
                                re.IGNORECASE | re.DOTALL)

    def strip_quotes(self, content):
        """
        Unquote given rule.

        Args:
            content (str): An import rule.

        Raises:
            InvalidImportRule: Raise exception if the rule is badly quoted
                (not started or not ended quotes).

        Returns:
            string: The given rule unquoted.
        """
        error_msg = "Following rule is badly quoted: {}"
        if (content.startswith('"') and content.endswith('"')) or \
           (content.startswith("'") and content.endswith("'")):
            return content[1:-1]
        # Quote starting but not ended
        elif (content.startswith('"') and not content.endswith('"')) or \
             (content.startswith("'") and not content.endswith("'")):
            raise InvalidImportRule(error_msg.format(content))
        # Quote ending but not started
        elif (not content.startswith('"') and content.endswith('"')) or \
             (not content.startswith("'") and content.endswith("'")):
            raise InvalidImportRule(error_msg.format(content))

        return content

    def remove_comments(self, content):
        """
        Remove all comment kind (inline and multiline) from given content.

        Args:
            content (str): A SCSS source.

        Returns:
            string: Given SCSS source with all comments removed.
        """
        return self.REGEX_COMMENTS.sub("", content)

    def filter_rules(self, path):
        """
        Lambda to filter items that:
        * Starts with http:// or https:// (this for external load only)
        * Ends with ".css" (they are not intended to be compiled)
        """
        return not(
            path.startswith("http://") or path.startswith("https://") or
            path.endswith(".css")
        )

    def flatten_rules(self, declarations):
        """
        Flatten returned import rules from regex.

        Because import rules can contains multiple items in the same rule
        (called multiline import rule), the regex ``REGEX_IMPORT_RULE``
        return a list of unquoted items for each rule.

        Args:
            declarations (list): A SCSS source.

        Returns:
            list: Given SCSS source with all comments removed.
        """
        rules = []

        for protocole, paths in declarations:
            # If there is a protocole (like 'url), drop it
            if protocole:
                continue
            # Unquote and possibly split multiple rule in the same declaration
            rules.extend([self.strip_quotes(v.strip())
                          for v in paths.split(',')])

        return list(filter(self.filter_rules, rules))

    def parse(self, content):
        """
        Parse a stylesheet document with a regex (``REGEX_IMPORT_RULE``)
        to extract all import rules and return them.

        Args:
            content (str): A SCSS source.

        Returns:
            list: Finded paths in import rules.
        """
        # Remove all comments before searching for import rules, to not catch
        # commented breaked import rules
        declarations = self.REGEX_IMPORT_RULE.findall(
            self.remove_comments(content)
        )
        return self.flatten_rules(declarations)


class SassImportsParser(ScssImportsParser):
    """
    Sass indented syntax parser to find ``@import`` rules.

    Multiline comments are not supported, it is actually too difficult to
    manage since they don't have a closing pattern ``*/`` and depend only
    on indentation continuation.

    Attributes:
        REGEX_IMPORT_RULE: Compiled regex used to find ``@import`` rules.
    """
    syntax = "sass"
    REGEX_IMPORT_RULE = re.compile(r'@import\s*(url)?\s*\(?([^;]+?)\)?(?:\n|$)',
                                   re.IGNORECASE)



# -*- coding: utf-8 -*-
"""
.. _Sass partials Reference:
    http://sass-lang.com/documentation/file.SASS_REFERENCE.html#partials

Finder
======

Finder is in charge to find *main Sass stylesheets* files to compile to CSS
files, meaning it will ignore all partials Sass stylesheets (see
`Sass partials Reference`_).

"""
import fnmatch
import os

from .exceptions import FinderException


def paths_by_depth(paths):
    """Sort list of paths by number of directories in it

    .. todo::

        check if a final '/' is consistently given or ommitted.

    :param iterable paths: iterable containing paths (str)
    :rtype: list
    """
    return sorted(
            paths,
            key=lambda path: path.count(os.path.sep),
            reverse=True
    )


class ScssFinder(object):
    """
    Project finder for SCSS sources

    Attributes:
        FINDER_STYLESHEET_EXTS: List of file extensions regarded as
            compilable stylesheet sources.
    """
    FINDER_STYLESHEET_EXTS = ["scss", "sass"]

    def get_relative_from_paths(self, filepath, paths):
        """
        Find the relative filepath from the most relevant multiple paths.

        This is somewhat like a ``os.path.relpath(path[, start])`` but where
        ``start`` is a list. The most relevant item from ``paths`` will be used
        to apply the relative transform.

        Args:
            filepath (str): Path to transform to relative.
            paths (list): List of absolute paths to use to find and remove the
                start path from ``filepath`` argument. If there is multiple
                path starting with the same directories, the biggest will
                match.

        Raises:
            boussole.exception.FinderException: If no ``filepath`` start could
                be finded.

        Returns:
            str: Relative filepath where the start coming from ``paths`` is
                removed.
        """
        for systempath in paths_by_depth(paths):
            if filepath.startswith(systempath):
                return os.path.relpath(filepath, systempath)

        raise FinderException("'Finder.get_relative_from_paths()' could not "
                              "find filepath start from '{}'".format(filepath))

    def is_partial(self, filepath):
        """
        Check if file is a Sass partial source (see
        `Sass partials Reference`_).

        Args:
            filepath (str): A file path. Can be absolute, relative or just a
                filename.

        Returns:
            bool: True if file is a partial source, else False.
        """
        path, filename = os.path.split(filepath)
        return filename.startswith("_")

    def is_allowed(self, filepath, excludes=[]):
        """
        Check from exclude patterns if a relative filepath is allowed

        Args:
            filepath (str): A relative file path. (exclude patterns are
                allways based from the source directory).

        Keyword Arguments:
            excludes (list): A list of excluding (glob) patterns. If filepath
                matchs one of patterns, filepath is not allowed.

        Raises:
            boussole.exception.FinderException: If given filepath is absolute.

        Returns:
            str: Filepath with new extension.
        """
        if os.path.isabs(filepath):
            raise FinderException("'Finder.is_allowed()' only accept relative"
                                  " filepath")

        if excludes:
            for pattern in excludes:
                if fnmatch.fnmatch(filepath, pattern):
                    return False
        return True

    def match_conditions(self, filepath, sourcedir=None, nopartial=True,
                         exclude_patterns=[], excluded_libdirs=[]):
        """
        Find if a filepath match all required conditions.

        Available conditions are (in order):

        * Is allowed file extension;
        * Is a partial source;
        * Is from an excluded directory;
        * Is matching an exclude pattern;

        Args:
            filepath (str): Absolute filepath to match against conditions.

        Keyword Arguments:
            sourcedir (str or None): Absolute sources directory path. Can be
                ``None`` but then the exclude_patterns won't be matched against
                (because this method require to distinguish source dir from lib
                dirs).
            nopartial (bool): Accept partial sources if ``False``. Default is
                ``True`` (partial sources fail matchind condition). See
                ``Finder.is_partial()``.
            exclude_patterns (list): List of glob patterns, if filepath match
                one these pattern, it wont match conditions. See
                ``Finder.is_allowed()``.
            excluded_libdirs (list): A list of directory to match against
                filepath, if filepath starts with one them, it won't
                match condtions.

        Returns:
            bool: ``True`` if match all conditions, else ``False``.
        """
        # Ensure libdirs ends with / to avoid missmatching with
        # 'startswith' usage
        excluded_libdirs = [os.path.join(d, "") for d in excluded_libdirs]

        # Match an filename extension admitted as compilable stylesheet
        filename, ext = os.path.splitext(filepath)
        ext = ext[1:]
        if ext not in self.FINDER_STYLESHEET_EXTS:
            return False

        # Not a partial source
        if nopartial and self.is_partial(filepath):
            return False

        # Not in an excluded directory
        if any(
            filepath.startswith(excluded_path)
            for excluded_path in paths_by_depth(excluded_libdirs)
        ):
            return False

        # Not matching an exclude pattern
        if sourcedir and exclude_patterns:
            candidates = [sourcedir]+excluded_libdirs
            relative_path = self.get_relative_from_paths(filepath, candidates)
            if not self.is_allowed(relative_path, excludes=exclude_patterns):
                return False

        return True

    def change_extension(self, filepath, new_extension):
        """
        Change final filename extension.

        Args:
            filepath (str): A file path (relative or absolute).
                new_extension (str): New extension name (without leading dot) to
                apply.

        Returns:
            str: Filepath with new extension.
        """
        filename, ext = os.path.splitext(filepath)
        return '.'.join([filename, new_extension])

    def get_destination(self, filepath, targetdir=None):
        """
        Return destination path from given source file path.

        Destination is allways a file with extension ``.css``.

        Args:
            filepath (str): A file path. The path is allways relative to
                sources directory. If not relative, ``targetdir`` won't be
                joined.
            absolute (bool): If given will be added at beginning of file
                path.

        Returns:
            str: Destination filepath.
        """
        dst = self.change_extension(filepath, "css")
        if targetdir:
            dst = os.path.join(targetdir, dst)
        return dst

    def compilable_sources(self, sourcedir, absolute=False, recursive=True,
                           excludes=[]):
        """
        Find all scss sources that should be compiled, aka all sources that
        are not "partials" Sass sources.

        Args:
            sourcedir (str): Directory path to scan.

        Keyword Arguments:
            absolute (bool): Returned paths will be absolute using
                ``sourcedir`` argument (if True), else return relative paths.
            recursive (bool): Switch to enabled recursive finding (if True).
                Default to True.
            excludes (list): A list of excluding patterns (glob patterns).
                Patterns are matched against the relative filepath (from its
                sourcedir).

        Returns:
            list: List of source paths.
        """
        filepaths = []

        for root, dirs, files in os.walk(sourcedir):
            # Sort structure to avoid arbitrary order
            dirs.sort()
            files.sort()
            for item in files:
                # Store relative directory but drop it if at root (".")
                relative_dir = os.path.relpath(root, sourcedir)
                if relative_dir == ".":
                    relative_dir = ""

                # Matching all conditions
                absolute_filepath = os.path.join(root, item)
                conditions = {
                    "sourcedir": sourcedir,
                    "nopartial": True,
                    "exclude_patterns": excludes,
                    "excluded_libdirs": [],
                }
                if self.match_conditions(absolute_filepath, **conditions):
                    relative_filepath = os.path.join(relative_dir, item)

                    if absolute:
                        filepath = absolute_filepath
                    else:
                        filepath = relative_filepath

                    filepaths.append(filepath)

            # For non recursive usage, break from the first entry
            if not recursive:
                break

        return filepaths

    def mirror_sources(self, sourcedir, targetdir=None, recursive=True,
                       excludes=[]):
        """
        Mirroring compilable sources filepaths to their targets.

        Args:
            sourcedir (str): Directory path to scan.

        Keyword Arguments:
            absolute (bool): Returned paths will be absolute using
                ``sourcedir`` argument (if True), else return relative paths.
            recursive (bool): Switch to enabled recursive finding (if True).
                Default to True.
            excludes (list): A list of excluding patterns (glob patterns).
                Patterns are matched against the relative filepath (from its
                sourcedir).

        Returns:
            list: A list of pairs ``(source, target)``. Where ``target`` is the
            ``source`` path but renamed with ``.css`` extension. Relative
            directory from source dir is left unchanged but if given,
            returned paths will be absolute (using ``sourcedir`` for
            sources and ``targetdir`` for targets).
        """
        sources = self.compilable_sources(
            sourcedir,
            absolute=False,
            recursive=recursive,
            excludes=excludes
        )
        maplist = []

        for filepath in sources:
            src = filepath
            dst = self.get_destination(src, targetdir=targetdir)

            # In absolute mode
            if targetdir:
                src = os.path.join(sourcedir, src)

            maplist.append((src, dst))

        return maplist



"""
Project management
==================
"""
import os

from .exceptions import SettingsInvalidError, SettingsBackendError
from .conf.json_backend import SettingsBackendJson
from .conf.yaml_backend import SettingsBackendYaml


class ProjectBase(object):
    """
    Project base

    Arguments:
        backend_name (string): Backend name, can be either ``json`` or
            ``yaml``.

    Attributes:
        _engines: Available Configuration backends. Read only.
        backend_name: Backend name to use from available ones.
        backend_engine: Backend engine selected from given name.
    """
    _engines = {
        "json": SettingsBackendJson,
        "yaml": SettingsBackendYaml,
    }

    def __init__(self, backend_name, **kwargs):
        self.backend_name = backend_name
        self.backend_engine = self.get_backend_engine(self.backend_name,
                                                      **kwargs)

    def get_backend_engine(self, name, **kwargs):
        """
        Get backend engine from given name.

        Args:
            (string): Path to validate.

        Raises:
            boussole.exceptions.SettingsBackendError: If given backend name
                does not match any available engine.

        Returns:
            object: Instance of selected backend engine.
        """
        if name not in self._engines:
            msg = "Given settings backend is unknowed: {}"
            raise SettingsBackendError(msg.format(name))

        return self._engines[name](**kwargs)


class ProjectStarter(ProjectBase):
    """
    Provide methods to create a new Sass Project
    """
    def valid_paths(self, *args):
        """
        Validate that given paths are not the same.

        Args:
            (string): Path to validate.

        Raises:
            boussole.exceptions.SettingsInvalidError: If there is more than one
                occurence of the same path.

        Returns:
            bool: ``True`` if paths are validated.
        """
        for i, path in enumerate(args, start=0):
            cp = list(args)
            current = cp.pop(i)
            if current in cp:
                raise SettingsInvalidError(
                    "Multiple occurences finded for path: {}".format(current)
                )

        return True

    def expand(self, basedir, config, sourcedir, targetdir, cwd):
        """
        Validate that given paths are not the same.

        Args:
            basedir (string): Project base directory used to prepend relative
                paths. If empty or equal to '.', it will be filled with current
                directory path.
            config (string): Settings file path.
            sourcedir (string): Source directory path.
            targetdir (string): Compiled files target directory path.
            cwd (string): Current directory path to prepend base dir if empty.

        Returns:
            tuple: Expanded arguments in the same order
        """

        # Expand home directory if any
        expanded_basedir = os.path.expanduser(basedir)
        expanded_config = os.path.expanduser(config)
        expanded_sourcedir = os.path.expanduser(sourcedir)
        expanded_targetdir = os.path.expanduser(targetdir)

        # If not absolute, base dir is prepended with current directory
        if not os.path.isabs(expanded_basedir):
            expanded_basedir = os.path.join(cwd, expanded_basedir)
        # Prepend paths with base dir if they are not allready absolute
        if not os.path.isabs(expanded_config):
            expanded_config = os.path.join(expanded_basedir,
                                           expanded_config)
        if not os.path.isabs(expanded_sourcedir):
            expanded_sourcedir = os.path.join(expanded_basedir,
                                              expanded_sourcedir)
        if not os.path.isabs(expanded_targetdir):
            expanded_targetdir = os.path.join(expanded_basedir,
                                              expanded_targetdir)

        # Normalize paths
        expanded_basedir = os.path.normpath(expanded_basedir)
        expanded_config = os.path.normpath(expanded_config)
        expanded_sourcedir = os.path.normpath(expanded_sourcedir)
        expanded_targetdir = os.path.normpath(expanded_targetdir)

        return (expanded_basedir, expanded_config, expanded_sourcedir,
                expanded_targetdir)

    def commit(self, sourcedir, targetdir, abs_config, abs_sourcedir,
               abs_targetdir):
        """
        Commit project structure and configuration file

        Args:
            sourcedir (string): Source directory path.
            targetdir (string): Compiled files target directory path.
            abs_config (string): Configuration file absolute path.
            abs_sourcedir (string): ``sourcedir`` expanded as absolute path.
            abs_targetdir (string): ``targetdir`` expanded as absolute path.
        """
        config_path, config_filename = os.path.split(abs_config)

        if not os.path.exists(config_path):
            os.makedirs(config_path)
        if not os.path.exists(abs_sourcedir):
            os.makedirs(abs_sourcedir)
        if not os.path.exists(abs_targetdir):
            os.makedirs(abs_targetdir)

        # Dump settings file
        self.backend_engine.dump({
            "SOURCES_PATH": sourcedir,
            "TARGET_PATH": targetdir,
            "LIBRARY_PATHS": [],
            "OUTPUT_STYLES": "nested",
            "SOURCE_COMMENTS": False,
            "EXCLUDES": []
        }, abs_config, indent=4)

    def init(self, basedir, config, sourcedir, targetdir, cwd='', commit=True):
        """
        Init project structure and configuration from given arguments

        Args:
            basedir (string): Project base directory used to prepend relative
                paths. If empty or equal to '.', it will be filled with current
                directory path.
            config (string): Settings file path.
            sourcedir (string): Source directory path.
            targetdir (string): Compiled files target directory path.

        Keyword Arguments:
            cwd (string): Current directory path to prepend base dir if empty.
            commit (bool): If ``False``, directory structure and settings file
                won't be created.

        Returns:
            dict: A dict containing expanded given paths.
        """
        if not basedir:
            basedir = "."

        # Expand home directory if any
        abs_basedir, abs_config, abs_sourcedir, abs_targetdir = self.expand(
            basedir, config,
            sourcedir, targetdir,
            cwd
        )

        # Valid every paths are ok
        self.valid_paths(abs_config, abs_sourcedir, abs_targetdir)

        # Create required directory structure
        if commit:
            self.commit(sourcedir, targetdir, abs_config, abs_sourcedir,
                        abs_targetdir)

        return {
            "basedir": abs_basedir,
            "config": abs_config,
            "sourcedir": abs_sourcedir,
            "targetdir": abs_targetdir,
        }



# -*- coding: utf-8 -*-
"""
Settings backend post processing
================================

Post processing methods are used to modify or validate given settings items.

Base backend inherit from ``SettingsPostProcessor`` to be able to use it in
its ``clean()`` method.
"""
import os

from ..exceptions import SettingsInvalidError

from . import SETTINGS_MANIFEST


class SettingsPostProcessor(object):
    """
    Mixin object for all available post processing methods to use in
    settings manifest (default manifest comes from ``SETTINGS_MANIFEST``).
    """
    settings_manifesto = SETTINGS_MANIFEST
    projectdir = ""

    def post_process(self, settings):
        """
        Perform post processing methods on settings according to their
        definition in manifest.

        Post process methods are implemented in their own method that have the
        same signature:

        * Get arguments: Current settings, item name and item value;
        * Return item value possibly patched;

        Args:
            settings (dict): Loaded settings.

        Returns:
            dict: Settings object possibly modified (depending from applied
            post processing).

        """
        for k in settings:
            # Search for post process rules for setting in manifest
            if (
                k in self.settings_manifesto and
                self.settings_manifesto[k].get("postprocess", None) is not None
            ):
                rules = self.settings_manifesto[k]["postprocess"]

                # Chain post process rules from each setting
                for method_name in rules:
                    settings[k] = getattr(self, method_name)(
                        settings,
                        k,
                        settings[k]
                    )

        return settings

    def _patch_expand_path(self, settings, name, value):
        """
        Patch a path to expand home directory and make absolute path.

        Args:
            settings (dict): Current settings.
            name (str): Setting name.
            value (str): Path to patch.

        Returns:
            str: Patched path to an absolute path.

        """
        if os.path.isabs(value):
            return os.path.normpath(value)

        # Expand home directory if any
        value = os.path.expanduser(value)

        # If the path is not yet an absolute directory, make it so from base
        # directory if not empty
        if not os.path.isabs(value) and self.projectdir:
            value = os.path.join(self.projectdir, value)

        return os.path.normpath(value)

    def _patch_expand_paths(self, settings, name, value):
        """
        Apply ``SettingsPostProcessor._patch_expand_path`` to each element in
        list.

        Args:
            settings (dict): Current settings.
            name (str): Setting name.
            value (list): List of paths to patch.

        Returns:
            list: Patched path list to an absolute path.

        """
        return [self._patch_expand_path(settings, name, item)
                for item in value]

    def _validate_path(self, settings, name, value):
        """
        Validate path exists

        Args:
            settings (dict): Current settings.
            name (str): Setting name.
            value (str): Path to validate.

        Raises:
            boussole.exceptions.SettingsInvalidError: If path does not exists.

        Returns:
            str: Validated path.

        """
        if not os.path.exists(value):
            msg = "Path from setting '{name}' does not exists: {value}"
            raise SettingsInvalidError(msg.format(name=name, value=value))

        return value

    def _validate_paths(self, settings, name, value):
        """
        Apply ``SettingsPostProcessor._validate_path`` to each element in
        list.

        Args:
            settings (dict): Current settings.
            name (str): Setting name.
            value (list): List of paths to patch.

        Raises:
            boussole.exceptions.SettingsInvalidError: Once a path does not
                exists.

        Returns:
            list: Validated paths.

        """
        return [self._validate_path(settings, name, item)
                for item in value]

    def _validate_required(self, settings, name, value):
        """
        Validate a required setting (value can not be empty)

        Args:
            settings (dict): Current settings.
            name (str): Setting name.
            value (str): Required value to validate.

        Raises:
            boussole.exceptions.SettingsInvalidError: If value is empty.

        Returns:
            str: Validated value.

        """
        if not value:
            msg = "Required value from setting '{name}' must not be empty."
            raise SettingsInvalidError(msg.format(name=name))

        return value



# -*- coding: utf-8 -*-
"""
Inspector
=========

Inspector is in charge to inspect a project about Sass stylesheets to search
for their dependencies.
"""
import io
import os

from collections import defaultdict

from .exceptions import CircularImport
from .parser import ScssImportsParser, SassImportsParser
from .resolver import ImportPathsResolver


class ScssInspector(ImportPathsResolver, ScssImportsParser):
    """
    Project inspector for SCSS sources

    Inspector is stateful, meaning you will need to invoke ``reset()`` then
    ``inspect()`` each time a project change, else the parents and children
    maps will be eventually incorrects.

    ``__init__`` method use ``reset`` method to initialize some internal
    buffers.

    Attributes:
        _CHILDREN_MAP: Dictionnary of finded direct children for each
            inspected sources.
        _PARENTS_MAP: Dictionnary of finded direct parents for each inspected
            sources.
        parsers: Dictionnary of available Sass format parsers, where key is the
            file extension related to the format and value is a parser
            instance.
    """
    parsers = {
        "scss": ScssImportsParser(),
        "sass": SassImportsParser(),
    }

    def __init__(self, *args, **kwargs):
        self.reset()

    def get_parser(self, path):
        """
        Returns parser depending of extension from given file path.

        Arguments:
            path (str): Path to split to find extension which will select the
                right parser. If file path does not have any extension or does not
                match any enabled extension, ``scss`` will be assumed on default.

        Returns:
            parser: Either ``ScssImportsParser`` or ``SassImportsParser``
            instance depending of filepath.
        """
        filepath, ext = os.path.splitext(path)

        # Remove leading dot from splitext to be able to match to
        # ``self.parsers`` keys
        if ext and ext.startswith("."):
            ext = ext[1:]

        if not ext or ext not in self.parsers:
            ext = "scss"

        return self.parsers.get(ext)

    def reset(self):
        """
        Reset internal buffers ``_CHILDREN_MAP`` and ``_PARENTS_MAP``.
        """
        self._CHILDREN_MAP = {}
        self._PARENTS_MAP = defaultdict(set)

    def look_source(self, sourcepath, library_paths=None):
        """
        Open a SCSS file (sourcepath) and find all involved files from
        import rules.

        This will fill internal buffers ``_CHILDREN_MAP`` and ``_PARENTS_MAP``.

        Arguments:
            sourcepath (str): Source file path to start searching for imports.

        Keyword Arguments:
            library_paths (list): List of directory paths for libraries to
                resolve paths if resolving fails on the base source path.
                Default to None.
        """
        # Don't inspect again source that has allready be inspected as a
        # children of a previous source
        if sourcepath not in self._CHILDREN_MAP:
            parser = self.get_parser(sourcepath)
            with io.open(sourcepath, "r", encoding="utf-8") as fp:
                finded_paths = parser.parse(fp.read())

            children = self.resolve(sourcepath, finded_paths,
                                    library_paths=library_paths)

            # Those files that are imported by the sourcepath
            self._CHILDREN_MAP[sourcepath] = children

            # Those files that import the sourcepath
            for p in children:
                self._PARENTS_MAP[p].add(sourcepath)

            # Start recursive finding through each resolved path that has not
            # been collected yet
            for path in children:
                if path not in self._CHILDREN_MAP:
                    self.look_source(path, library_paths=library_paths)

        return

    def inspect(self, *args, **kwargs):
        """
        Recursively inspect all given SCSS files to find imported dependencies.

        This does not return anything. Just fill internal buffers about
        inspected files.

        Note:
            This will ignore orphan files (files that are not imported from
            any of given SCSS files).

        Arguments:
            *args: One or multiple arguments, each one for a source file path
                to inspect.

        Keyword Arguments:
            library_paths (list): List of directory paths for libraries to
                resolve paths if resolving fails on the base source path.
                Default to None.
        """
        library_paths = kwargs.get("library_paths", None)

        for sourcepath in args:
            self.look_source(sourcepath, library_paths=library_paths)

    def _get_recursive_dependancies(self, dependencies_map, sourcepath,
                                    recursive=True):
        """
        Return all dependencies of a source, recursively searching through its
        dependencies.

        This is a common method used by ``children`` and ``parents`` methods.

        Arguments:
            dependencies_map (dict): Internal buffer (internal buffers
                ``_CHILDREN_MAP`` or ``_PARENTS_MAP``) to use for searching.
            sourcepath (str): Source file path to start searching for
                dependencies.

        Keyword Arguments:
            recursive (bool): Switch to enable recursive finding (if True).
                Default to True.

        Raises:
            CircularImport: If circular error is detected from a source.

        Returns:
            set: List of dependencies paths.
        """
        # Direct dependencies
        collected = set([])
        collected.update(dependencies_map.get(sourcepath, []))

        # Sequence of 'dependencies_map' items to explore
        sequence = collected.copy()
        # Exploration list
        walkthrough = []

        # Recursive search starting from direct dependencies
        if recursive:
            while True:
                if not sequence:
                    break
                item = sequence.pop()

                # Add current source to the explorated source list
                walkthrough.append(item)

                # Current item children
                current_item_dependancies = dependencies_map.get(item, [])

                for dependency in current_item_dependancies:
                    # Allready visited item, ignore and continue to the new
                    # item
                    if dependency in walkthrough:
                        continue
                    # Unvisited item yet, add its children to dependencies and
                    # item to explore
                    else:
                        collected.add(dependency)
                        sequence.add(dependency)

                # Sourcepath has allready been visited but present itself
                # again, assume it's a circular import
                if sourcepath in walkthrough:
                    msg = "A circular import has occured by '{}'"
                    raise CircularImport(msg.format(current_item_dependancies))

                # No more item to explore, break loop
                if not sequence:
                    break

        return collected

    def children(self, sourcepath, recursive=True):
        """
        Recursively find all children that are imported from the given source
        path.

        Arguments:
            sourcepath (str): Source file path to search for.

        Keyword Arguments:
            recursive (bool): Switch to enabled recursive finding (if True).
                Default to True.

        Returns:
            set: List of finded parents path.
        """
        return self._get_recursive_dependancies(
            self._CHILDREN_MAP,
            sourcepath,
            recursive=True
        )

    def parents(self, sourcepath, recursive=True):
        """
        Recursively find all parents that import the given source path.

        Arguments:
            sourcepath (str): Source file path to search for.

        Keyword Arguments:
            recursive (bool): Switch to enabled recursive finding (if True).
                Default to True.

        Returns:
            set: List of finded parents path.
        """
        return self._get_recursive_dependancies(
            self._PARENTS_MAP,
            sourcepath,
            recursive=True
        )



# -*- coding: utf-8 -*-
"""
Settings model
==============

This define the model object containing settings that will be passed to
interfaces.

"""
import copy

from . import SETTINGS_MANIFEST


# Default values for initial settings object
DEFAULT_SETTINGS = {k: copy.deepcopy(v["default"])
                    for k, v in SETTINGS_MANIFEST.items()}


class Settings(object):
    """
    Settings model object

    Class init method fills object attributes from default settings
    (``DEFAULT_SETTINGS``) then update it with initial settings if given.

    Settings are available as object attributes, there is also a private
    ``_settings`` attribute containing a dict of all stored settings. You are
    strongly advised to never directly manipulate the ``_settings`` attribute.
    Instead, allways use the ``update()`` method.

    Note:
        Model is only about data model, there is no other validation that
        available 'fields' from ``DEFAULT_SETTINGS``.

        If you intend to manually open and fill a Settings instance, remember
        to allways use absolute paths in your settings. Relative path will
        cause issues in resolving that lead to wrong compilations.

        You may also apply post processor validation to ensure your datas.

    Keyword Arguments:
        initial (dict): A dictionnary of settings for initial values.
    """
    def __init__(self, initial={}):
        self._settings = copy.deepcopy(DEFAULT_SETTINGS)
        if initial:
            initial = self.clean(initial)
            self._settings.update(initial)
        self.set_settings(self._settings)

    def clean(self, settings):
        """
        Filter given settings to keep only key names available in
        ``DEFAULT_SETTINGS``.

        Args:
            settings (dict): Loaded settings.

        Returns:
            dict: Settings object filtered.

        """
        return {k: v for k, v in settings.items() if k in DEFAULT_SETTINGS}

    def set_settings(self, settings):
        """
        Set every given settings as object attributes.

        Args:
            settings (dict): Dictionnary of settings.

        """
        for k, v in settings.items():
            setattr(self, k, v)

    def update(self, settings):
        """
        Update object attributes from given settings

        Args:
            settings (dict): Dictionnary of elements to update settings.

        Returns:
            dict: Dictionnary of all current saved settings.

        """
        settings = self.clean(settings)

        # Update internal dict
        self._settings.update(settings)

        # Push every setting items as class object attributes
        self.set_settings(settings)

        return self._settings



# -*- coding: utf-8 -*-
import click

from sass import __version__ as libsasspython_version

from _sass import libsass_version

from boussole import __version__


@click.command()
@click.pass_context
def version_command(context):
    """
    Print out version information.
    """
    click.echo("Boussole {} (libsass-python {}) (libsass {})".format(
        __version__,
        libsasspython_version,
        libsass_version,
    ))



"""
Logging
=======
"""
import logging

import colorlog


def init_logger(level, printout=True):
    """
    Initialize app logger to configure its level/handler/formatter/etc..

    Todo:
        * A mean to raise click.Abort or sys.exit when CRITICAL is used;

    Args:
        level (str): Level name (``debug``, ``info``, etc..).

    Keyword Arguments:
        printout (bool): If False, logs will never be outputed.

    Returns:
        logging.Logger: Application logger.
    """
    root_logger = logging.getLogger("boussole")
    root_logger.setLevel(level)

    # Redirect outputs to the void space, mostly for usage within unittests
    if not printout:
        from io import StringIO
        dummystream = StringIO()
        handler = logging.StreamHandler(dummystream)
    # Standard output with colored messages
    else:
        handler = logging.StreamHandler()
        handler.setFormatter(
            colorlog.ColoredFormatter(
                "%(asctime)s - %(log_color)s%(message)s",
                datefmt="%H:%M:%S"
            )
        )

    root_logger.addHandler(handler)

    return root_logger



# -*- coding: utf-8 -*-
"""
Project configuration
=====================

Boussole work on per project configurations stored in a settings file for each
project.

Backends behavior is to search for a settings file in the given directory, read
it, possibly patch its values and then return a
:class:`boussole.conf.model.Settings object`.

Almost all paths in settings will be expanded to absolute paths if they are
not allready so:

* If the path start with a home directory character, the home directory is used
  to expand the path;
* If the path is relative, expand it to absolute using directory from settings
  file location;

Also note, that Sass files from libraries directories are never compiled.
"""
# Manifest define default values and post process methods for each setting
# Warning: Order does matter on "postprocess" methods
SETTINGS_MANIFEST = {
    "LIBRARY_PATHS": {
        "default": [],
        "postprocess": (
            "_patch_expand_paths",
            "_validate_paths",
        ),
    },
    "SOURCES_PATH": {
        "default": None,
        "postprocess": (
            "_validate_required",
            "_patch_expand_path",
            "_validate_path",
        ),
    },
    "TARGET_PATH": {
        "default": None,
        "postprocess": (
            "_validate_required",
            "_patch_expand_path",
            "_validate_path",
        ),
    },
    "OUTPUT_STYLES": {
        "default": "nested",
        "postprocess": [],
    },
    "SOURCE_COMMENTS": {
        "default": False,
        "postprocess": [],
    },
    "SOURCE_MAP": {
        "default": False,
        "postprocess": [],
    },
    "EXCLUDES": {
        "default": [],
        "postprocess": [],
    },
}
